# -*- coding: utf-8 -*-
"""
Created on Fri May 13 22:43:05 2022

@author: Jean
"""

# pipeline如果传入多个货劳名称列表则是并行模式，CPU与GPU版都支持。
# hanlp 2.1.0b27，HanLP 2.1仍处于beta状态，各beta版本间会有变化，
# 比如自定义词典的格式与使用方式，依存句法分析与语义依存分析的输出等。
import hanlp
# 用于挂载用户自定义词典分词，
# 发票货物劳务名称中有大量的专有名词，并不常见于训练用的语料库中，导致分词不够准确，需要发票专用的自定义词典。
from hanlp.components.tokenizers.transformer import TransformerTaggingTokenizer
# 用于为上面的用户自定义词典作词性标注，提取货物劳务名称需要标注词性，货物劳务名称显然都是名词，这是个名词短语。
from hanlp.components.mtl.tasks.pos import TransformerTagging
import pandas as pd
import time, random, math
import json,re
import collections
from igraph import *
# 装入后面用到的预训练模型，PyTorch上的NLP深度学习预训练模型， Pytorch 1.11.0 + cuda 11.3。
# 第一次调用时会下载到~/.hanlp之下，大的有几百M，一共要几G。
# 中文分词模型
# COARSE_ELECTRA_SMALL_ZH预训练模型的效果没有FINE_ELECTRA_SMALL_ZH好，前者是粗粒度，后者是细粒度
# LARGE_ALBERT_BASE预训练模型是TensorFlow模型，Pytorch与TensorFlow模型一起运行有点问题，后面在详细了解。
# tokenizer = hanlp.load(hanlp.pretrained.tok.COARSE_ELECTRA_SMALL_ZH)
t1 = time.time()
tokenizer: TransformerTaggingTokenizer = hanlp.load(hanlp.pretrained.tok.FINE_ELECTRA_SMALL_ZH)
# 词性标注模型
tagger: TransformerTagging = hanlp.load(hanlp.pretrained.pos.CTB9_POS_ELECTRA_SMALL)
# 依存句法分析模型
# 自动下载模型到本地:  https://file.hankcs.com/hanlp/dep/ctb9_dep_electra_small_20220216_100306.zip
# 然后下载引用的transformer模型：https://huggingface.co/models/hfl/chinese-electra-180g-small-discriminator/config.json
# 自己手工下载transformer模型存储到 %HANLP_HOME/transformers/hlf/chinese-electra-180g-small-discriminator
# 然后更改 %HANLP_HOME/dep/ctb9_dep_electra_small_20220216_100306/config.json,
# 改引用模型的路径为：  "transformer": "D:/HanLP/transformers/hlf/chinese-electra-180g-small-discriminator"
# 然后就会从本地加载transformer模型，也加快了加载的速度。
# 参阅： https://www.youtube.com/watch?v=bnaCKP6mLZI
syntactic_parser = hanlp.load(hanlp.pretrained.dep.CTB9_DEP_ELECTRA_SMALL)
# 语义依存分析模型
# 下载模型:  https://file.hankcs.com/hanlp/sdp/semeval16_sdp_electra_small_20220218_140116.zip
# transformer模型处理同上。
semantic_parser = hanlp.load(hanlp.pretrained.sdp.SEMEVAL16_ALL_ELECTRA_SMALL_ZH)
# 命名实体识别模型
ner = hanlp.load(hanlp.pretrained.ner.MSRA_NER_ELECTRA_SMALL_ZH)
t2 = time.time()
print(t2-t1)

# 货物劳务名称，测试用
text = ["利尔康牌95%酒精消毒液", "汤臣倍健番茄红素维生素E软胶囊", "95号汽油"]

# 中文分词
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/tok_stl.ipynb
han_words = tokenizer(text)
print(han_words)

# 词性标注
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/pos_stl.ipynb
han_pos = tagger(han_words)
print(han_pos)

# 中文分词
# 挂载用户自定义词典，强制模式
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/demo_custom_dict_stl.py
tokenizer.dict_force = None
han_words = tokenizer(text)
print(han_words)
# 分词字典是个集合
tokenizer.dict_force = {'汤臣倍健', '番茄红素', '维生素E'}
han_words = tokenizer(text)
print(han_words)

# 词性标注
# 挂载用户自定义词典，强制模式
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/demo_pos_dict.py
tagger.dict_tags=None
han_pos = tagger(han_words)
print(han_pos)
# 词性标注字典是个字典
tagger.dict_tags={'汤臣倍健': 'NN', '番茄红素': 'NN', '维生素E': 'NN'}
han_pos = tagger(han_words)
print(han_pos)

# 依存句法分析
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/dep_stl.ipynb
trees = syntactic_parser(han_words)
print(trees)
trees = syntactic_parser(han_words, conll=False)
print(trees)
from hanlp_common.document import Document
for words,tree in zip(han_words, trees):
    doc = Document(tok = words, dep = tree)
    doc.pretty_print()

# 语义依存分析
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/sdp_stl.ipynb
han_sp = semantic_parser(han_words)
print(han_sp)

# 命名实体识别
# https://github.com/hankcs/HanLP/blob/doc-zh/plugins/hanlp_demo/hanlp_demo/zh/ner_stl.ipynb
print(ner(han_words))

# pipeline模式已挂载了用户自定义词典
pipeline = hanlp.pipeline() \
    .append(tokenizer, output_key='tokens') \
    .append(tagger, input_key='tokens', output_key='part_of_speech_tags') \
    .append(syntactic_parser, input_key='tokens', output_key='syntactic_dependencies') \
    .append(semantic_parser, input_key='tokens', output_key='semantic_dependencies')
docs = pipeline(text)
print(docs)

# 自定义函数， 返回依存句法分析json结果字符串，抽取key子集
# 传入的text是个货劳名称列表
def syntacticParse(parser, text):
    res = parser(text)
    # items = res.syntactic_dependencies # HanLP 2.0 -->2.1
    sentences = res["syntactic_dependencies"]
    poses = res["part_of_speech_tags"]
    reStrs = []
    for sentence, pos in zip(sentences, poses):
        items2 = {}; items2.setdefault("items",[])
        for word, ps in zip(sentence, pos):
            item2 ={key:value for key ,value in word.items() \
                        # if key in ["cpos","deprel","form","head","id"]}    # HanLP 2.0 -->2.1   
                        if key in ["deprel","form","head","id"]}
            item2["upos"] = ps
            # print(item2)
            items2.get("items").append(item2) 
        reStrs.append(json.dumps(items2,ensure_ascii=False))
    return reStrs

# 测试依存句法分析函数
reStrs = syntacticParse(pipeline, text)
print(reStrs)

# 建立语法树（图）函数
def getTree(text):
    res = json.loads(text)
    try:
        items = res["items"]
    except KeyError:
        print(text)
        return None
    ids = []; pos = []; words = []; head = []; deprel = []
    for item in items:
        # ids.append(item["id"]); words.append(item["form"]); pos.append(item["cpos"])    # HanLP 2.0 -->2.1     
        ids.append(item["id"]); words.append(item["form"]); pos.append(item["upos"])
        head.append(item["head"]); deprel.append(item["deprel"])
            
    vs = pd.DataFrame({"id":ids,"word":words,"postag":pos})
    es = pd.DataFrame({"id":ids,"head":head,"deprel":deprel})
    vd = vs.to_dict(orient='records')
    ed = es.to_dict(orient='records')
    g = Graph.DictList(vd,ed, vertex_name_attr="id",edge_foreign_keys=('id', 'head'),directed=True)
    # 给根节点赋非空值
    vs = g.vs.select(id=0)   
    vs["word"] = ""; vs["postag"] = ""
    g.vs["label"]=[str(v["id"])+":"+v["word"]+"\n"+v["postag"] for v in g.vs]
    g.es["label"]=g.es["deprel"]
    return g

# 画语法树
g = getTree(reStrs[2])
random.seed(42)
plot(g,bbox=(400,300))


# 自定义函数， 返回语义依存分析json结果字符串，抽取key子集
# 传入的text是个货劳名称列表
def semanticParse(parser, text):
    res = parser(text)
    print(res)
    # items = res.syntactic_dependencies # HanLP 2.0 -->2.1
    sentences = res["semantic_dependencies"]
    poses = res["part_of_speech_tags"]
    reStrs = []
    for sentence, pos in zip(sentences, poses):
        items2 = {}; items2.setdefault("items",[])
        for word, ps in zip(sentence, pos):
            item2 ={key:value for key ,value in word.items() \
                    # if key in ["cpos","deprel","form","head","id"]}    # HanLP 2.0 -->2.1   
                    if key in ["deps","form","id"]}
            item2["upos"] = ps
            #print(item2)
            items2.get("items").append(item2)        
        reStrs.append(json.dumps(items2,ensure_ascii=False))
    return reStrs


reStrs2 = semanticParse(pipeline, text)
print(reStrs2)


# 建立语义网络（图）函数， 语义分析结果为一个有向无环图，称为语义依存图（Semantic Dependency Graph）。
# 不过测试例子显示返回结果中存在环路。
def getTree2(text):
    res = json.loads(text)
    try:
        items = res["items"]
    except KeyError:
        print(text)
        return None
    ids = []; pos = []; words = []
    source = []; target=[]; reltype = []
    for item in items:
        # ids.append(item["id"]); words.append(item["form"]); pos.append(item["cpos"])  # HanLP2.0 -->2.1      
        ids.append(item["id"]); words.append(item["form"]); pos.append(item["upos"])
        # 展开语义图的边
        for link in item["deps"]:
            source.append(item["id"])
            target.append(link[0])
            reltype.append(link[1])
            if link[0]==0:                    
                # 补充根结点, 并给属性赋非空值
                ids.append(0); words.append(""); pos.append("")
                    
    vs = pd.DataFrame({"id":ids,"word":words,"postag":pos})
    es = pd.DataFrame({"id":source,"head":target,"deps":reltype})
    vd = vs.to_dict(orient='records')
    ed = es.to_dict(orient='records')
    g = Graph.DictList(vd,ed, vertex_name_attr="id",edge_foreign_keys=('id', 'head'),directed=True)
    g.vs["label"]=[str(v["id"])+":"+v["word"]+"\n"+v["postag"] for v in g.vs]
    g.es["label"]=g.es["deps"]
    return g

# 画语义依存图， 语义分析结果为一个有向无环图，称为语义依存图（Semantic Dependency Graph）。
g2 = getTree2(reStrs2[0])
random.seed(42)
plot(g2,bbox=(400,300))


# 根据业务上的观察与统计规则，遍历图并输出名词的顺序（对包含的名词重新排序）。
# 从根结点开始遍历语法树，取最先出现的名词作货劳名称，所以层级高的优先，同层级的在后面的优先
def get_hlmc(g, verbose = False):
    # 建立邻接表
    adj = [[n.index for n in v.neighbors("in")] for v in g.vs]
    spcialWords = re.compile("[\.|\-|%]") #特殊名称， "."软件版本号， "-"具体型号，跳过
    spcialPuncs = ["-", "/", "*", "+"]       #特殊标点符号，标示里面包含并列关系,
    skipWords = re.compile("[顶|装]")
    # 广度优先搜索遍历语法树
    def traverse(adj, paths, depth):
        # 记录遍历的深度
        depth += 1;  nxt_paths = []
        # 遍历所有的路径
        for path in paths['paths']:
            v = g.vs[path[-1]]
            # 如果不是并列词，则反转邻接表，改成同层级的后面的优先，排在后面的名词优先，
            # 发票语法结构的特点，前面的名词多数是修饰语         
            if v["word"] not in spcialPuncs and depth>1:
                adj[path[-1]].reverse()
            for idx in adj[path[-1]]:
                v1 = g.vs[idx]
                # 如 汤臣倍健多种维生素咀嚼片-迪士尼漫威装, 根结点是 漫威装
                if v1["word"] in spcialPuncs:
                    adj[path[-1]].reverse()  # 子结点中有并列关系，邻接表恢复为正常顺序。                    
                    break
            # 遍历路径中最后一个结点的所有邻居
            for nxt in adj[path[-1]]:
                v = g.vs[nxt]
                if verbose:
                    print(depth,v["word"],v["postag"])
                if "N"  in v["postag"]:   # 如果是名词，则加入候选货劳清单中
                    flag = True
                    if spcialWords.findall(v["word"]) and v["postag"]!= "NM" :  # 软件版本号，具体型号，跳过，药名除外。
                        flag = False
                    if v["postag"]== "NT":  # 品牌，跳过
                        flag = False
                    matched = skipWords.search(v["word"])   # 检查是否包含要跳过的特殊名词
                    if matched:
                        if matched.span()[0]+1 == len(v["word"]):   # 漫威装 在最后，跳过；安装费 在中间，接受
                            flag = False
                    if flag:
                        paths["hlmc"].append(v["word"])
                        paths["pos"].append(v["postag"])
                nxt_path = path + [nxt]   # 在路径中加入下一个结点，生成新的路径
                nxt_paths.append(nxt_path)
        # 路径的深度增加了一级
        paths['paths'] = nxt_paths                            
        if len(nxt_paths)==0:
            # 没有新路径了
            return paths
        else:
            # 往路径深度递归调用一级
            return traverse(adj, paths, depth)
    # 从根结点开始递归遍历语法树
    v = g.vs.select(id=0)[0]
    # 返回候选货劳名称列表，排在前面的在语法树中的层级较高，同层级中则排在较后的优先级较高
    res = traverse(adj, {'paths': [[v.index]], 'hlmc': [], "pos": []}, 0)
    # 没有找到名词，返回一级结点作货劳名称
    if len(res["hlmc"]) == 0:
        v1 = v.neighbors("in")[0]
        res["hlmc"].append(v1["word"]); res["pos"].append(v1["postag"])
    
    return res

#  测试遍历语法树得到货劳名称
hlmc1 = get_hlmc(g, True)
print(hlmc1["hlmc"][0])

#  测试遍历语法树得到货劳名称
hlmc2 = get_hlmc(g2, True)
print(hlmc2["hlmc"][0])

# 从CSV文件装入用户自定义词典
types = {"word":str,"pos":str}
dicZdy = pd.read_csv('D:/temp/data/增值税自开票2017/词汇导入demo-HanLP.csv', sep=',', header=None, dtype=types, encoding="GBK")
# 词性标注字典是个字典
dictPos = dict(collections.OrderedDict(zip(dicZdy.iloc[:,0],dicZdy.iloc[:,1])))
# 分词字典是个集合
dicTag = set(dicZdy.iloc[:,0])
tokenizer.dict_force = dicTag
han_words = tokenizer(text)
print(han_words)
tagger.dict_tags= dictPos
print(tagger(han_words))


# 句法依存分析或语义依存分析前对汇总后的货物劳务名称做预处理
def preprocess(text):     
    # 去掉前后空格等特殊处理
    text = text.strip()
    if text == "." or text == "/"  or text == "-":
        text = ""    
    # 去掉几种括号及其中内容，如果去掉后没有内容了，则予以保留
    text1 = text
    text = re.sub("\\(.*?\\)|（.*?）", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\[.*?\\]|【.*?】", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\{.*?\\}|｛.*?｝", "", text)
    if len(text) == 0:
        text = text1
    # 去掉商品及货物分类简称
    text1 = text        
    text = re.sub("\\*.*?\\*", "", text)
    if len(text) == 0:
        text = text1
    # 去掉单边括号后的内容，这种情况不少
    text1 = text        
    text = text.split("（")[0]  #	722 维达（Vinda 	822	荣耀8全网通（FRD 	825	彩电挂式安装调试卡（
    if len(text) == 0:
        text = text1
    text1 = text                
    text = text.split("）")[0] 
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split("(")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split(")")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text
    text = re.sub("及", " ", text)  # 国际快件费及附加 -> 国际快件费 附加
    if len(text) == 0:
        text = text1        
    text1 = text                
    text = text.split(" ")[0]  # 空格分隔的多个并列，只留第一个，如 "小米5C 移动版 3"
    if len(text) == 0:
        text = text1        
    text = re.sub("[0-9]L", "", text)  # 胶瓶美汁源果粒橙12*1.25L，去掉L，引起分词错误 
    
    return text


print(preprocess("胶瓶美汁源果粒橙12*1.25L"))
# 货物劳务名称，测试用
text = ["利尔康牌95%酒精消毒液", "汤臣倍健番茄红素维生素E软胶囊", "95号汽油(国IV)"]
text2 = [(lambda x:preprocess(x))(y) for y in text]


